//
// Copyright (C) 1991 Texas Instruments Incorporated.
//
// Permission is granted to any individual or institution to use, copy, modify,
// and distribute this software, provided that this complete copyright and
// permission notice is maintained, intact, in all copies and supporting
// documentation.
//
// Texas Instruments Incorporated provides this software "as is" without
// express or implied warranty.
//
// Created: MJF 03/27/89 -- Initial design and implementation.
// Updated: MJF 04/15/89 -- Implemented Base list class.
// Updated: JCB 06/05/89 -- Fixed merge and sort.
// Updated: JCB 06/20/89 -- Implemented next_union, next_intersection,
//                          next_difference, next_xor.
// Updated: JCB 06/21/89 -- Modified next() and prev() to check for a current
//                          position not in THIS list, i.e. !traversal
// Updated: MJF 08/03/89 -- Improved operator== by checking for identical nodes
// Updated: MJF 08/10/89 -- Changed return values of operator+=, etc to List ref
// Updated: MJF 08/11/89 -- Changed methods which return new list references:
//                          tail, last, but_last, member, sublist and copy
// Updated: LGO 09/07/89 -- Made next, reference and dereference inline
// Updated: LGO 09/07/89 -- Moved the guts of find to the template classes.
// Updated: LGO 09/08/89 -- use do_find instead of compare_data in most places.
// Updated: LGO 09/08/89 -- Move push to the template classes.
// Updated: MBN 09/20/89 -- Added conditional exception handling
// Updated: LGO 10/05/89 -- Don't sort lists with less than 2 elements
// Updated: MBN 10/11/89 -- Change "current_position" to "curpos" and also
//                          "previous_position" to "prevpos"
// Updated: LGO 12/04/89 -- Make binary set operators not inline
// Updated: LGO 12/05/89 -- Make sort/merge predicate ANSI compatable
// Updated: MJF 03/12/90 -- Added group names to RAISE
// Updated: MJF 05/22/90 -- Fixed remove_duplicates
// Updated:  VDN 02/21/92 -- new lite version and fix memory leaks
//
// This  file contains member and friend  function implementation code for  the
// CoolList class defined in the  Base_List.h header  file.  Where appropriate  and
// possible, interfaces  to, and  us of,  existing system   functions has  been
// incorporated.  An overview of the structure of the CoolList class, along  with a
// synopsis of each member and friend function, can be found in the Base_List.h
// header file.

#ifndef BASE_LISTH				// If CoolList class not defined,
#include <cool/Base_List.h>			// include header file
#endif

CoolList NIL = CoolList();			// NIL -- Global NIL CoolList

// CoolList_Node() -- Constructor
// Input:         None
// Output:        None

CoolList_Node::CoolList_Node() {}		// Constructor


// ~CoolList_Node -- Destructor (not inline because it's virtual)
//                   Virtual is needed to call ~CoolList_Node<Type>.
// Input:         None
// Output:        None

CoolList_Node::~CoolList_Node() {;}


// get_data()  -- Gets data of node. There is no data for CoolList class.
// Input:         None.
// Output:        A void* pointer.

void* CoolList_Node::get_data() {
  cout << "\nWarning:  CoolList_Node::get_data() has been called.\n";
  return NULL;
}


// set_data()  -- Sets data of node to specified value. There is not data
//                to set for CoolList class.
// Input:         A void* pointer.
// Output:        None.

void CoolList_Node::set_data(void*) {
#if ERROR_CHECKING
  //RAISE (Warning, SYM(CoolList_Node), SYM(Redundant_Method_Call),
  printf ("CoolList_Node::set_data(): Method called is redundant.\n");
#endif
}


// ~CoolList -- Destructor (not inline because it's virtual)
// Input:         None
// Output:        None

CoolList::~CoolList () {;}

// Prev() -- decrement current position. If NULL, set it to last.
// Input:    None.
// Output:   TRUE or FALSE.

Boolean CoolList::prev() {
  register CoolList_Node* cp = this->curpos;
  CoolList_Node* prev = this->prevpos;
  // if cp invalid (i.e. NULL) the following test will fail
  if (prev != NULL && prev->next == cp) {
    // already have the previous position
    cp = prev;
    this->prevpos = NULL; // this isn't needed, but why not...
  } else {
    // find previous node of current position
    // When cp is invalid, this finds the last node.
    register CoolList_Node* np = this->node_ptr;
    prev = NULL;
    if (np != NULL) {
      while(np->next != cp)
	prev = np, np = np->next;
      cp = np;
      this->prevpos = prev;    
    }
  }
  // current position not found in CoolList
  return (this->curpos = cp) != NULL;
}

// operator== -- Returns TRUE if similar data in THIS and the specified CoolList.
//               Uses the static CoolList comparer function on elements.
// Input:        A CoolList reference.
// Output:       TRUE or FALSE.

Boolean CoolList::operator==(const CoolList& l) const {
  if (this == &l) return TRUE;  // same CoolLists
  for (CoolList_Node *np = this->node_ptr, *np_l = l.node_ptr;
      np != NULL && np_l != NULL; 
      np = np->next, np_l = np_l->next) {
    if (np == np_l)
      return TRUE;   // same node pointers
    else if (!this->compare_data(np->get_data(), np_l->get_data()))
      return FALSE;
  }
  if (np == NULL && np_l == NULL) return TRUE;
  else return FALSE;
}


// tail() -- Sets CoolList passed to the the nth tail of this CoolList. With no 2nd
//           argument, tail() sets CoolList passed to the 1st tail of this CoolList.
// Input:    A CoolList reference to store the nth tail elements of THIS and
//           the positive integer, n. (default 1)
// Output:   None.

void CoolList::tail(CoolList& l, int n) {
  this->dereference(l.node_ptr);
  int count = n;
  CoolList_Node* np;
  for (np = this->node_ptr; np != NULL && count > 0; np = np->next, count--);
  // if n < length, return nth tail
  if (np != NULL && count == 0) {
    this->curpos = np;
    this->reference(np);
    l.node_ptr = np;
  }
  else l.node_ptr = NULL;
}


// last() -- Sets CoolList passed to the n last elements of this CoolList. With no
//           2nd argument, sets CoolList passed to the last element of thie CoolList.
// Input:    A CoolList reference to store the n last elements of THIS and
//           the positive interger, n (default 1).
// Output:   None.

void CoolList::last(CoolList& l, int n) {
  // last n elements = (size-n)th tail
  int nth = this->length() - n;   
  if (nth >= 0)
    tail(l, nth);
  else {
    this->dereference(l.node_ptr);
    l.node_ptr = NULL;
  }
}


// but_last() -- Sets CoolList passed to all but the last n elements of THIS List. 
//               With no 2nd argument, sets CoolList passed to all but the last
//               element
// Input:        A positive integer (default 1).
// Output:       An altered THIS with all but the n last elements.

void CoolList::but_last(CoolList& l, int n) {
  this->reset();				// Current position invalid
  CoolList_Node* np = this->node_ptr;
  this->dereference(l.node_ptr);		// Delete nodes in l
  if (n == 0) {
    this->reference(np);			
    l.node_ptr = np;				// Points at all nodes of THIS
  }
  else {
    // get the number of nodes in this CoolList and copy into CoolList l
    int no_nodes = this->length() - n; 
    if (no_nodes > 0 && np != NULL)  {
      CoolList_Node* first_np = this->insert_after_node(np->get_data(), NULL);
      CoolList_Node* rest_np = first_np;	
      no_nodes--;
      for(np = np->next; 
	  np != NULL && no_nodes > 0; 
	  np = np->next, no_nodes--) {
	rest_np = this->insert_after_node(np->get_data(), rest_np);
      }
      l.node_ptr = first_np;			// Create totally new nodes
    } else
      l.node_ptr = NULL;
  }
}


// clear() -- Removes all nodes of THIS CoolList.
// Input:     None.
// Output:    None.

void CoolList::clear() {
  this->reset();				// make current position invalid   
  this->dereference(this->node_ptr);		// Delete nodes in this.
  this->node_ptr = NULL;
}


// length() -- Returns the number of elements in THIS CoolList.
// Input:      None.
// Output:     An interger representing the number of elements in THIS.

int CoolList::length() {
  int count = 0;
  for (CoolList_Node* np = this->node_ptr; np != NULL; np = np->next, count++);
  return count;
}


// position -- Returns current position.
// Input:      THIS.
// Output:     An integer representing current position.

int CoolList::position() {
  if (this->curpos == NULL) return -1;
  int index = 0;
  for (CoolList_Node* np = this->node_ptr;  np != NULL; np = np->next, index++)
    if (np == this->curpos) return index;
  return -1;
}

// search() -- Returns TRUE if the specified CoolList is a subset of this CoolList
// Input:      A reference to subList to be searched.
// Output:     TRUE or FALSE.

Boolean CoolList::search(const CoolList& l) {
  CoolList_Node* tnode = this->node_ptr;
  CoolList_Node* lnode = l.node_ptr;
  if (tnode == lnode) {
    // this CoolList and subList l have same head node pointers
    this->curpos = tnode;
    this->prevpos = NULL;
    return TRUE;  
  }
  if (lnode == NULL) return FALSE;	// subList l is nil
  while(tnode != NULL &&
	this->do_find(tnode, lnode->get_data(), this->curpos, this->prevpos)) {
    tnode = this->curpos;
    if (tnode == lnode) {
      // a node in this CoolList and head node of CoolList l are same
      return TRUE;
    }
    // data of node in this CoolList is also in node of CoolList l
    // continue searching rest of data
    for (CoolList_Node *np2 = tnode->next, *lnp = lnode->next; 
	 np2 != NULL && lnp != NULL;
	 np2 = np2->next, lnp = lnp->next) {
      if (np2 == lnp) {
	// node in this CoolList and node in CoolList l are same
	return TRUE;
      }
      if  (!this->compare_data(np2->get_data(), lnp->get_data()))
	break;
    }
    if (lnp == NULL) {
      // found data of CoolList l in this CoolList
      return TRUE;
    }
    tnode = tnode->next;
  } 
  return FALSE;
}


// sublist() -- Returns TRUE if THIS CoolList contains second argument and sets
//              first argument to a sublist in THIS CoolList starting at the 1st
//              occurence of second argument.
// Input:       A reference to the CoolList which will be set to a sublist within
//              THIS CoolList; and a reference to the CoolList in search of.
// Output:      TRUE or FALSE.

Boolean CoolList::sublist(CoolList& l, const CoolList& subl) {
  this->dereference(l.node_ptr);
  if (this->search(subl)) {
    this->reference(this->curpos);
    l.node_ptr = this->curpos;  
    return TRUE;
  } else {
    l.node_ptr = NULL;
    return FALSE;
  }
}


// copy() -- Mutates *this to store a copy of elements in specified List
// Input:    List containing elements to be copied from
// Output:   None.

void CoolList::copy (const CoolList& l) {
  cout << "****warning: copy from argument into *this****" << endl;
  this->reset();				// make current position invalid
  this->dereference(this->node_ptr);		// free current nodes of *this
  CoolList_Node* np = l.copy_nodes();		// make new list of nodes of l
  this->node_ptr = np;				// all with ref_count = 1
}

// reverse() -- Reverses the order of the elements of THIS CoolList.
// Input:       None.
// Output:      None.

void CoolList::reverse() {
  this->reset();				// Current position invalid
  CoolList_Node *np, *np2, *rev_head;
  rev_head = NULL;
  for (np = this->node_ptr; np != NULL; np = np2) {
    np2 = np->next;
    np->next = rev_head;
    rev_head = np;
  }
  this->node_ptr = rev_head;
}


// prepend() -- Prepends the specified CoolList to the front of this CoolList and
//              returns TRUE if specified CoolList is not NULL
// Input:       A reference to the CoolList to be prepended.
// Output:      Always TRUE.

Boolean CoolList::prepend(const CoolList& l) {
  // first get copy of nodes in CoolList l 
  CoolList_Node* lnode = l.copy_nodes();
  if (lnode == NULL)
    return FALSE;
  // now prepend copy of l to front of THIS CoolList.  
  this->curpos = lnode;
  this->prevpos = NULL;
  CoolList_Node* last_np = this->curpos;
  if (last_np != NULL) {  
    // find last node of new copy of l
    while(last_np->next != NULL) last_np = last_np->next;
    // insert last node of CoolList l in front of head node of this CoolList
    // and set head node of this CoolList to head node of CoolList l
    last_np->next = this->node_ptr;
    this->node_ptr = this->curpos;
  }
  return TRUE;
}


// append() -- Appends the specified CoolList to the end of this CoolList and returns
//             TRUE if specified CoolList is not NULL.
// Input:      A reference to the CoolList to be appended.
// Output:     TRUE or FALSE.

Boolean CoolList::append(const CoolList& l) {
  CoolList_Node* lnode = l.node_ptr;
  if (lnode == NULL)
    return FALSE;
  this->curpos = lnode;				// curpos is head node of l
  CoolList_Node* last_np = this->node_ptr;
  if (last_np == NULL) { 
    // there are no nodes in this;
    // just set head node of this CoolList to head node of CoolList l
    this->node_ptr = this->curpos;
    this->prevpos = NULL;
  }
  else {
    // find last node of this
    while(last_np->next != NULL) last_np = last_np->next;
    // add head node of CoolList l after last node of this CoolList
    last_np->next = this->curpos;
    this->prevpos = last_np;
  }
  this->reference(this->curpos);
  return TRUE;
}


// set_tail() -- Sets the nth tail of THIS CoolList to the specified CoolList
// Input:        A CoolList reference, and a count (default 1).
// Output:       TRUE if nth tail exists, FALSE otherwise.

Boolean CoolList::set_tail(const CoolList& l, int n) {
  // find node at start of nth tail
  int count = n;
  for (CoolList_Node *np = this->node_ptr, *prev_np = NULL; 
      np != NULL && count > 0;
      prev_np = np, np = np->next, count--);
  if (np != NULL && count == 0) {
     this->curpos = l.node_ptr;	// Current position=head node
     this->prevpos = prev_np;		// Previous position is nth-1
     // replace nth node with head node of CoolList l
     if (prev_np == NULL) this->node_ptr = this->curpos; // when n=0
     else prev_np->next = this->curpos;	       // when n>0
     this->reference(this->curpos);
     // removing nth tail
     this->dereference(np);
     return TRUE;
   }
  return FALSE;
}


// remove_duplicates() -- Removes all duplicate elements in THIS CoolList.
// Input:                 None.
// Output:                TRUE if at least one item is removed, FALSE otherwise.

Boolean CoolList::remove_duplicates() {
  Boolean success = FALSE;			// success flag
  CoolList_Node* np = this->node_ptr;
  CoolList_Node* np_next;
  while ((np != NULL) && ((np_next = np->next) != NULL)) {
    while (this->do_find (np_next, np->get_data(), this->curpos, this->prevpos))
      {
	this->dereference(this->remove());	// duplicate nodes go to trash
	success = TRUE;
	if ((np_next = this->curpos) == NULL) break;
      }
    np = np->next;
  }
  return success;
}


// set_intersection() -- Changes THIS list to contain everything that is an element
//                   of both this and the specified CoolList.
// Input:            A reference to CoolList.
// Output:           None.

void CoolList::set_intersection(const CoolList& l) {
  this->reset();				// make current position invalid
  if (l.node_ptr == NULL)
    this->clear();
  else {
    CoolList_Node* np = this->node_ptr;
    CoolList_Node* prev_np = NULL;
    CoolList_Node* next_np;
    while(np != NULL) {
      CoolList_Node* cp;
      CoolList_Node* pp;
      if (!l.do_find(l.node_ptr, np->get_data(), cp, pp)) {
        next_np = np->next;
        if (prev_np == NULL) 
	  this->node_ptr = np->next;		// next node is at head
        else 
	  prev_np->next = np->next;		// next node is in middle
        this->reference(np->next);
        this->dereference(np);			// delete np from this CoolList.
        np = next_np;				// set to next node to look at
      }
      else {
        prev_np = np; 
        np = np->next;
      }
    }
  }
}


// set_union() -- Changes THIS list to contain everything that is an element of
//             either THIS or the specified CoolList
// Input:      A reference to CoolList.
// Output:     None.

void CoolList::set_union(const CoolList& l) {
  this->reset();				// make current position invalid
  CoolList_Node* tnode = this->node_ptr;
  CoolList_Node* lnode = l.node_ptr;
  if (tnode ==  NULL && lnode != NULL) { 
    this->node_ptr = lnode;			// this is empty, so set head node 
    this->reference(lnode);			// of this to head node of l
  }
  else
    for (; lnode !=NULL; lnode = lnode->next) {
      CoolList_Node* cp;
      CoolList_Node* pp;
      if (!do_find(this->node_ptr, lnode->get_data(), cp, pp))
	this->push(lnode->get_data());		// push new data of l into this.
    }
}


// set_difference() -- Removes from THIS CoolList the elements which also appears in
//                 the specified CoolList.
// Input:          A reference to CoolList.
// Output:         None.

void CoolList::set_difference(const CoolList& l) {
  this->reset();				// make current position invalid
  if (l.node_ptr != NULL) {
    CoolList_Node* np = this->node_ptr;
    CoolList_Node* prev_np = NULL;
    CoolList_Node* next_np;
    // if node in this CoolList is found in CoolList l, remove it from this CoolList
    while (np != NULL) {
      CoolList_Node* cp;
      CoolList_Node* pp;
      if (l.do_find(l.node_ptr, np->get_data(), cp, pp)) {
        next_np = np->next;
        if (prev_np == NULL) 
	  this->node_ptr = np->next;		// Next node is at head
        else 
	  prev_np->next = np->next;		// next node is in middle
        this->reference(np->next);
        this->dereference(np);			// delete np from this CoolList.
        np = next_np;				// set to next node to look at
      }
      else {
        prev_np = np; 
        np = np->next;
      }
    }
  }
}


// set_xor() -- Changes CoolList to contain those elements which appear in
//                   exactly one of THIS CoolList and the specified CoolList.
// Input:            A reference to CoolList.
// Output:           None.

void CoolList::set_xor(const CoolList& l) {
  this->reset();	// make current position invalid
  // first get copy of nodes in CoolList l
  CoolList* lcopy = this->new_list((CoolList_Node*)NULL);
  lcopy->copy(l);
  // remove elements in this CoolList from copy of CoolList l
  lcopy->set_difference(*this);
  // remove elements in CoolList l from this CoolList  
  this->set_difference(l);
  // append elements left in lcopy to end of this CoolList
  this->append(*lcopy);
  delete lcopy;					// destructor is virtual
}

// next_intersection() -- Sets current position to next item in the
//                        intersection of this CoolList and the specified CoolList.
// Input:                 None.
// Output:                TRUE if the next item exists,  FALSE otherwise.

Boolean CoolList::next_intersection(const CoolList& l) {
  CoolList_Node* np = this->curpos;
  if (!this->traversal) {			// end of intersection
    this->curpos = NULL;
    return FALSE;
  }
  if (np == NULL)				// starting fresh
    np = this->node_ptr;
  else
    np = np->next;				// increment position
  for (; np != NULL; np = np->next) {
    CoolList_Node* cp;
    CoolList_Node* pp;
    if (l.do_find(l.node_ptr, np->get_data(), cp, pp)) {
      this->curpos = np;
      return TRUE;
    }
  }
  this->traversal = FALSE;			// now traversing second CoolList
  return FALSE;
}


// next_union() -- Sets current position to next item in the union of this
//                  CoolList and the specified CoolList.
// Input:           None.
// Output:          TRUE if the next item exists,  FALSE otherwise.

Boolean CoolList::next_union(const CoolList& l) {
  CoolList_Node* np = this->curpos;
  if (np == NULL && !this->traversal)		// end of both CoolLists
    return FALSE;
  if (np == NULL && this->traversal)		// starting fresh
    np = this->node_ptr;
  else  
    if (np->next == NULL && this->traversal) {	// end of first CoolList
      this->traversal = FALSE;			// now traversing second CoolList
      np = l.node_ptr;
    }
  else
    if (!this->traversal)  np = np->next;	// increment position

  if (this->traversal) {			// still in this CoolList
    if (this->curpos != NULL) 
      this->curpos = np->next;	// increment current position
    else this->curpos = np;
    return TRUE;			
  }

  for (; np != NULL; np = np->next) {
    CoolList_Node* cp;
    CoolList_Node* pp;
    if (!do_find(this->node_ptr, np->get_data(), cp, pp)) {
      this->curpos = np;
      return TRUE;
    }
  }
  return FALSE;
}


// next_difference() -- Sets current position to next item in the difference
//                      of this CoolList and the specified CoolList
// Input:               None.
// Output:              TRUE if the next item exists,  FALSE otherwise.

Boolean CoolList::next_difference(const CoolList& l) {
  CoolList_Node* np = this->curpos;
  if (np == NULL && !this->traversal) {  	// end of this CoolList
    return FALSE;
  }
  if (np == NULL)				// starting fresh
    np = this->node_ptr;
  else {
    this->traversal = FALSE;			// now traversing second CoolList
    np = np->next;				// incrementing position
  }
  for (; np != NULL; np = np->next) {
    CoolList_Node* cp;
    CoolList_Node* pp;
    if (!l.do_find(l.node_ptr, np->get_data(), cp, pp)) {
      this->curpos = np;
      return TRUE;
    }
  }
  return FALSE;
}


// next_xor() -- Sets current position to next item in exclusive_or
//                        of this CoolList and the specified CoolList.
// Input:                 None.
// Output:                TRUE if the next item exists,  FALSE otherwise.

Boolean CoolList::next_xor(const CoolList& l) {
  CoolList_Node* cp;
  CoolList_Node* pp;
  CoolList_Node* np = this->curpos;
  if (!this->traversal && np == NULL)		// end of both CoolLists
    return FALSE;
  if (np == NULL && this->traversal)		// starting fresh
    np = this->node_ptr;
  else 
    if (np->next == NULL && this->traversal) {	// end of first CoolList
      this->traversal = FALSE;			// now traversing second CoolList
      np = l.node_ptr;
    }
  else
    np = np->next;				// increment position
  if (this->traversal) {			// still in this CoolList
    for (; np != NULL; np = np->next) 
      if (!l.do_find(l.node_ptr, np->get_data(), cp, pp)) {
        this->curpos = np;
	return TRUE;			
      } 
    this->traversal = FALSE;			// now traversing second CoolList
    np = l.node_ptr; 
  }
  for (; np != NULL; np = np->next)
    if (!do_find(this->node_ptr, np->get_data(), cp, pp)) {
      this->curpos = np;
      return TRUE;
    }
  return FALSE;
}


// describe() -- Describes internal structure of each node of this CoolList
// Input:        The output stream to display description on.
// Output:       None.

void CoolList::describe(ostream& os) {
  os << "\nCoolList " << this << ":\n";
  int count = 0;
  for (CoolList_Node* np = this->node_ptr; np != NULL; np = np->next, count++) 
  {
    os << "Node" << count << ":\n";
    os << " Data = ";
    this->output_data(os,np);
    os << "\n";
    os << " Ref count = " << np->ref_count << "\n";
  }
}


// operator<<() -- Overload output operator for CoolList objects
// Input:          An output stream reference and CoolList reference
// Output:         An output stream reference.

ostream& operator<<(ostream& os, const CoolList& l) {
  os << "(";
  CoolList_Node* np = l.node_ptr;
  if (np != NULL) {
    l.output_data(os, np);
    for (np = np->next; np != NULL; np = np->next) {
      os << " ";
      l.output_data(os,np);
    }
  }
  return os << ")";
}


// new_list() -- Returns new CoolList with head node initialized to specified node
// Input:        The node pointer.
// Output:       A pointer to the new CoolList.

CoolList* CoolList::new_list(CoolList_Node*) {
#if ERROR_CHECKING
  //RAISE Warning, SYM(CoolList_Node), SYM(Redundant_Method_Call),
  printf ("CoolList_Node::new_list(): Method called is redundant.\n");
#endif
  return (CoolList*)NULL;  
}


// insert_before_node() -- Inserts a new node before the specified node.
// Input:                  A Type reference and a Node pointer.
// Output:                 A pointer to the new node.

CoolList_Node* CoolList::insert_before_node(void*, CoolList_Node*) {
#if ERROR_CHECKING
  //RAISE Warning, SYM(CoolList_Node), SYM(Redundant_Method_Call),
  printf ("CoolList_Node::insert_before_node(): Method called is redundant.\n");
#endif
 return NULL;
}


// insert_after_node() -- Inserts a new node after the specified node.
// Input:                 A Type reference and a Node pointer.
// Output:                A pointer to the new node.

CoolList_Node* CoolList::insert_after_node(void*, CoolList_Node*) const {
#if ERROR_CHECKING
  //RAISE Warning, SYM(CoolList_Node), SYM(Redundant_Method_Call),
  printf ("CoolList_Node::insert_after_node(): Method called is redundant.\n");
#endif
 return NULL;
}


// compare_data() -- Compares data using default compare function of this CoolList
// Input:            Two void* pointers which will be type cast.
// Output:           None.

Boolean CoolList::compare_data(const void*, const void*) const {
#if ERROR_CHECKING
  //RAISE Warning, SYM(CoolList_Node), SYM(Redundant_Method_Call),
  printf ("CoolList_Node::compare_data(): Method called is redundant.\n");
#endif
  return FALSE;
}


// output_data()  -- Outputs node data from specified stream.
// Input:            An output stream reference and node pointer.
// Output:           A void* pointer of data.

void CoolList::output_data(ostream&, const CoolList_Node*) const {

#if ERROR_CHECKING
  //RAISE Warning, SYM(CoolList_Node), SYM(Redundant_Method_Call),
  printf ("CoolList_Node::output_data(): Method called is redundant.\n");
#endif
}


// copy_nodes() -- Returns a copy of nodes in THIS CoolList.
// Input:          NONE.
// Output:         The first node pointer of copy.

CoolList_Node* CoolList::copy_nodes() const {
  CoolList_Node *np, *first_np;
  if ((np = this->node_ptr) == NULL)  
     first_np = NULL;
  else {
    // copy first node of this CoolList
    first_np = this->insert_after_node(np->get_data(), NULL);
    CoolList_Node* rest_np = first_np;

    // copy rest of nodes of this CoolList
    for (np = np->next; np != NULL; np = np->next) {
      rest_np = this->insert_after_node(np->get_data(), rest_np);
    }
  }
  return first_np;
}


void CoolList::free_nodes(CoolList_Node* np) {
  do {
    if (np->ref_count < 0) {
      cout << "\nWarning: negative ref count of " << np->ref_count;
      cout << " for data ";
      this->output_data(cout, np);
      cout << " and ";
#if GENERIC_TYPECHECK
      cout << (this->type_of())->name() << " = ";
#endif
      cout << this << ".\n";
    } 
    // no longer sharing this node with other CoolLists
    CoolList_Node* next_np = np->next;     // save pointer to next node
    np->next = NULL;
//  delete np->data;			    // not nec. data is ptr or obj
    delete np;				    // delete node structure
    np = next_np;			    // check reference of next node
  }
  while (np != NULL && --(np->ref_count) <= 0);
}

// operator= -- Assigns THIS to the specified CoolList.
// Input:       A reference to a CoolList which THIS will be assigned to.
// Output:      The modified THIS.

CoolList* CoolList::operator=(const CoolList& l) {
  this->reset();				// make current position invalid
  this->reference(l.node_ptr);			// will point to new stuff
  this->dereference(this->node_ptr);		// delete old stuff
  this->node_ptr = l.node_ptr;
  return this;
}


// operator[]() -- Returns the nth node of this CoolList
// Input:          A positive integer index.
// Output:         A pointer to the nth node.

CoolList_Node* CoolList::operator[] (int n) {
  int count = n;
  for (CoolList_Node *np = this->node_ptr, *prev_np = NULL;
      np != NULL && count > 0;
      prev_np = np, np = np->next, count--);

  // if n<length,  found nth node
  if (np != NULL && count == 0) {
    this->curpos = np;
    this->prevpos = prev_np;
    return np;
  }
  else return (CoolList_Node*)NULL;  // or error
}


// position() -- Returns the position of the specified data item in this CoolList,
//               else returns -1 if not found
// Input:        A void* pointer of a data item.
// Output:       The integer position.

int CoolList::position(const void* x) {
  int index = 0;
  // NOTE: It's faster to call do_find, then count the position,
  // than to do the search ourselves using compare_data and get_data.
  if(!this->do_find(this->node_ptr, x, this->curpos, this->prevpos))
    return -1;
  else {
    int index = 0;
    CoolList_Node* cp = this->curpos;
    CoolList_Node* np = this->node_ptr;
    while (np != cp) np = np->next, index++;
    return index;
  }
}


// do_find() -- Returns TRUE if the specified element is a member of THIS CoolList.
// Input:       A void* pointer of data item to be searched.
// Output:      TRUE or FALSE.

Boolean CoolList::do_find(CoolList_Node*, const void*,
		      CoolList_Node*&, CoolList_Node*&) const {
  cout << "\nWarning:  CoolList_Node::find(x) has been called.\n";
  return FALSE;
}

// member() -- Returns TRUE if THIS CoolList contains the specified element and
//             sets the specified CoolList to a subList in THIS CoolList starting
//             at the first occurence of element.
// Input:      A reference to the CoolList which will be set to a subCoolList within
//             THIS CoolList, and a void* pointer of data item to be searched.
// Output:     A CoolList pointer to some tail of THIS.

Boolean CoolList::member(CoolList& l,const void* x) {
  this->dereference(l.node_ptr);
  if (this->do_find(this->node_ptr, x, this->curpos, this->prevpos)) {
    this->reference(this->curpos);
    l.node_ptr = this->curpos;
    return TRUE;
  } else {
    l.node_ptr = NULL;
    return FALSE;
  }
}


// push() -- Prepends the specified data item to the front of CoolList
// Input:    A void* pointer of the data item to be prepended.
// Output:   Always TRUE.

Boolean CoolList::push(const void* x) {
  this->curpos = this->insert_before_node(x, this->node_ptr);
  this->node_ptr = this->curpos;
  this->prevpos = NULL;
  return TRUE;
}


// push_new() -- Pushes a new element at head of THIS CoolList if it is not already
//               a member of the CoolList.
// Input:        A void* pointer of new data.
// Output:       TRUE if item not on CoolList, FALSE otherwise.

Boolean CoolList::push_new(const void* x) {
  CoolList_Node* cp;
  CoolList_Node* pp;
  if (!this->do_find(this->node_ptr, x, cp, pp)) // don't change curpos
    return push(x);
  else
    return FALSE;
}


// push_end() -- Appends the specified data item to the end of this CoolList
// Input:        A void* pointer of the data item to be appended.
// Output:       Always TRUE.

Boolean CoolList::push_end(const void* x) {
  CoolList_Node* last_np = this->node_ptr;
  if (last_np == NULL) {  
    // there arn't any nodes in this CoolList; x will be the head node
    this->node_ptr = this->curpos = this->insert_after_node(x, NULL);
    this->prevpos = NULL;
  }
  else {
    CoolList_Node* cp = this->curpos;
    // find last node of this CoolList
    if (cp != NULL && cp->next == NULL)
      last_np = cp;
    else
      while(last_np->next != NULL) last_np = last_np->next;
    // insert x after last node
    this->curpos = this->insert_after_node(x, last_np);  
    this->prevpos = last_np;
  }
 return TRUE;
}


// push_end_new() -- Pushes a new element at end of THIS CoolList if it is not
//                   already a member of the CoolList.
// Input:            A void* pointer of new data.
// Output:           TRUE if item not on CoolList, FALSE otherwise.

Boolean CoolList::push_end_new(const void* x) {
  if (!this->do_find(this->node_ptr, x, this->curpos, this->prevpos))
    return this->push_end(x);
  else
    return FALSE;
}


// pop() -- Removes and returns head element of THIS CoolList.
// Input:   None.
// Output:  Node pointer containing head data element of THIS CoolList.

CoolList_Node* CoolList::pop() {
  CoolList_Node* old_head = this->node_ptr;
  if (old_head != NULL) {
    this->reset();				// make current position invalid
    this->node_ptr = old_head->next;
    this->reference(old_head->next);
    return old_head;
  }
  else return NULL;
}


// remove() -- Removes item at current position.
// Input:      None.
// Output:     The node pointer of item removed.

CoolList_Node*  CoolList::remove() {
  if (this->curpos == NULL)
    return NULL;
  CoolList_Node* np = this->prevpos;
  CoolList_Node* cp = this->curpos;
  if (np == NULL || np->next != cp) { 
    np  = this->node_ptr;
    if (np == cp) {				// node to remove is head node
      this->curpos = np->next;
      this->node_ptr = np->next;		// this points to next of head
      this->reference(np->next);		
      return np;
    }
    // find previous node current position
    while (np != NULL && np->next != this->curpos) np = np->next;
    if (np == NULL) return NULL;		// error: curpos was not in list
    else this->prevpos = np;
  }
  np->next = cp->next;
  this->curpos = cp->next;
  this->reference(cp->next);
  return cp;
}


// remove() -- Removes the first occurence of the specified item in this CoolList
// Input:      A refernce to data item to be removed.
// Output:     TRUE if item found and removed, FALSE otherwise.

Boolean CoolList::remove(const void* x) {
  // find node for x
  if (this->do_find(this->node_ptr, x, this->curpos, this->prevpos)) {
    this->dereference(this->remove());		// throw node to trash, since 
    return TRUE;				// it is not returned
  }
  return FALSE;
}


// replace() -- Replaces the first occurence of specified data item in THIS
//              CoolList with a new value
// Input:       Two void* pointers.
// Output:      TRUE if item found and replaced, FALSE otherwise.

Boolean CoolList::replace(const void* old_data, const void* new_data) {
  // find node of old_data in this CoolList
  if(this->do_find(this->node_ptr, old_data, this->curpos, this->prevpos)) {
    // replace old_data in node with new_data
    this->curpos->set_data(new_data);
    return TRUE;
  }
  return FALSE;
}


// replace_all() -- Replaces all occurences of the specified data item in THIS
//                  CoolList with a new value.
// Input:           Two void* pointers.
// Output:          TRUE if at least one item found and replaced, else FALSE

Boolean CoolList::replace_all(const void* old_data, const void* new_data) {
  this->reset();	// make current position invalid
  Boolean success = FALSE; // return value
  CoolList_Node* pp;
  CoolList_Node* np = this->node_ptr;
  while (np != NULL && this->do_find(np, old_data, np, pp)) {
    np->set_data(new_data);
    success = TRUE;
    np = np->next;
  }
  return success;
}


// sort() -- Sorts the elements of THIS using the specified predicate function.
//           The predicate function returns TRUE if and only if the first
//           element preceeds the second. The sort routine uses the Sort
//           algorithm as given in "Numerical Recipes in C" p247.
// Input:    A predicate function pointer.
// Output:   None.

void CoolList::sort(Predicate f) {
  this->reset();	// make current position invalid
  CoolList_Node* np;
  int l, j, ir, i;
  int n = this->length();
  if (n < 2) return;	// No sense sorting if less than two elements
  CoolList_Node** node_array = new CoolList_Node*[n+1];
  // put the nodes of THIS into an array which will be used by the
  // heap sort algorithm
  for (np = this->node_ptr, i = 1; np != NULL; np = np->next, i++)
     node_array[i] = np;
  // the heap sort algorithm
  CoolList_Node* temp;
  l = (n >> 1) + 1;
  ir = n;
  while (TRUE) {
    if (l > 1)
      temp = node_array[--l];
    else {
      temp = node_array[ir];
      node_array[ir] = node_array[1];
      if (--ir == 1) {
	node_array[1] = temp;
	break;
      }
    }
    i = l;
    j = i << 1;
    while (j <= ir) {
      if (j < ir && (*f) ((node_array[j])->get_data(),
			  (node_array[j+1])->get_data()) < 0)
	++j;
      if ((*f) (temp->get_data(), (node_array[j])->get_data()) < 0) {
	node_array[i] = node_array[j];
	j += (i = j);
      }
      else
	j = ir + 1;
    }
    node_array[i] = temp;
  }

  // put the sorted nodes of the array back into THIS
  this->node_ptr = node_array[1];
  for (i = 1; i < n; i++) {
     (node_array[i])->next = node_array[i+1];
   }
  (node_array[n])->next = NULL;

  delete [] node_array;
}


// merge() -- Merges the elements of THIS CoolList with the elements of the
//            specified CoolList sorted with the specified predicate function.
//            If THIS is sorted already, then the result of the merge will be
//            sorted. Otherwise there are no guarantees for the sortedness of
//            the result.
// Input:     A reference to a CoolList to be merged and a predicate function 
//            pointer.
// Output:    None.

void CoolList::merge(const CoolList& l, Predicate f) {
  this->reset();	// make current position invalid
  CoolList_Node* tnode = this->node_ptr;
  if (tnode == NULL)
    // if this CoolList is NULL, just assign this CoolList to CoolList l
    this->operator=(l);
  else {
    // begin merging CoolList l into this CoolList
    CoolList_Node *lnode;

    // first merge nodes of CoolList l in front of first node of this CoolList
    // until a node in this CoolList is less than a node in CoolList l
    for (lnode = l.node_ptr;
	 lnode != NULL &&
	 (*f) (tnode->get_data(), lnode->get_data()) >= 0;
	 lnode = lnode->next) {
      tnode = this->insert_before_node(lnode->get_data(), tnode);
      this->node_ptr = tnode;
    }
      
    // merge rest of nodes in CoolList l into this CoolList
    CoolList_Node* tnode_prev = tnode;
    tnode = tnode->next;
    while (lnode != NULL) {
      if (tnode == NULL) {
	// if the end of THIS is reached, 
	// add the CoolList l nodes to the end of this CoolList
	for (; lnode !=NULL; lnode = lnode->next)
	  tnode_prev = this->insert_after_node(lnode->get_data(), tnode_prev);
      }
      else if ((*f)(tnode->get_data(), lnode->get_data()) < 0) {
	// if node in this CoolList is less than node in CoolList l
	// just go to next node in this CoolList
	tnode_prev = tnode;
	tnode = tnode->next;      
      } else {
	// if node in CoolList l is less than node in this CoolList
	// insert the CoolList l node into THIS
	// and goto next node in CoolList l
	tnode_prev->next = this->insert_before_node(lnode->get_data(), tnode);
	tnode_prev = tnode_prev->next;
	lnode = lnode->next;
      }
    }
  }
}


// insert_before() -- Inserts the specified item before the current position
// Input:             A void* pointer.
// Output:            TRUE if current position is valid, FALSE otherwise.

Boolean CoolList::insert_before(const void* new_item) {
  CoolList_Node* cp = this->curpos;
  CoolList_Node *np = this->prevpos;
  if (cp == NULL)
    return FALSE;
  if (np == NULL || np->next != cp) {
    np = this->node_ptr;
    // current position is head node    
    if (np == cp)
      return this->push(new_item);
    // find previous node of current position in CoolList
    while (np != NULL && np->next != cp) np = np->next;
    if (np == NULL)
      return FALSE;  // couldn't find current position
    else
      this->prevpos = np;
  }
  // insert new_item before current position and
  // set current position to this new node
  this->curpos = this->insert_before_node(new_item, cp);
  np->next = this->curpos;
  return TRUE;
}


// insert_after() -- Inserts the specified item after the current position
// Input:            A void* pointer.
// Output:           TRUE if current position is valid, FALSE otherwise.

Boolean CoolList::insert_after(const void* new_item) {
  if (this->curpos == NULL)
    return FALSE;
  this->prevpos = this->curpos;
  this->curpos = this->insert_after_node(new_item,
						   this->curpos);
  return TRUE;
}


// insert_before() -- Inserts the specified new item before the specified
//                    target item in this CoolList
// Input:             Two data void* pointers.
// Output:            TRUE if target item found, FALSE otherwise.

Boolean CoolList::insert_before(const void* new_item, const void* target_item) {
  // find node with target item
  if (this->do_find(this->node_ptr, target_item, this->curpos, this->prevpos)){
    CoolList_Node *np = this->curpos;
    CoolList_Node *prev_np = this->prevpos;
    // insert new item before target item
    this->curpos = this->insert_before_node(new_item, np);
    if (prev_np == NULL)
      this->node_ptr = this->curpos;	 // new item is the head node
    else
      prev_np->next = this->curpos; // new item is a  middle node
    return TRUE;
  }
  return FALSE;
}


// insert_after() -- Inserts the specified new item after the specified target
//                   item in this CoolList
// Input:            Two data void* pointers.
// Output:           TRUE if target item found, FALSE otherwise.

Boolean CoolList::insert_after(const void* new_item, const void* target_item) {
  // find node with target item
  CoolList_Node *np;
  if (this->do_find(this->node_ptr, target_item, np, this->prevpos)) {
    this->curpos = this->insert_after_node(new_item, np);
    this->prevpos = np;
    return TRUE;
  }       
  return FALSE;
}


// value_error -- Raise exception for CoolList::value() method
// Input:         Type string
// Output:        None

void CoolList::value_error (const char* Type) {
  //RAISE Error, SYM(CoolList), SYM(Invalid_Cpos),
  printf ("CoolList<%s>::value(): Invalid current position.\n", Type);
  abort ();
}


// get_error -- Raise exception for CoolList::get() method
// Input:       Type string
// Output:      None

void CoolList::get_error (const char* Type, int n) {
  //RAISE Error, SYM(CoolList), SYM(Negative_Index),
  printf ("CoolList<%s>::get(): Negative index %d.\n", Type, n);
  abort ();
}


// before_error -- Raise exception for CoolList::insert_before() method
// Input:          Type string
// Output:         None

void CoolList::before_error (const char* Type) {
  //RAISE Error, SYM(CoolList), SYM(Invalid_Cpos),
  printf ("CoolList<%s>::insert_before(): Invalid current position.\n", Type);
  abort ();
}


// after_error -- Raise exception for CoolList::insert_after() method
// Input:         Type string
// Output:        None

void CoolList::after_error (const char* Type) {
  //RAISE Error, SYM(CoolList), SYM(Invalid_Cpos),
  printf ("CoolList<%s>::insert_after(): Invalid current position.\n", Type);
  abort ();
}


// bracket_error -- Raise exception for CoolList::operator[]() method
// Input:           Type string
// Output:          None

void CoolList::bracket_error (const char* Type, int n) {
  //RAISE Error, SYM(CoolList), SYM(Negative_Index),
  printf ("CoolList<%s>::operator[](): Negative index %d.\n", Type, n);
  abort ();
}


// pop_error -- Raise exception for CoolList::pop() method
// Input:       Type string
// Output:      None

void CoolList::pop_error (const char* Type) {
  //RAISE Error, SYM(CoolList), SYM(No_Elements),
  printf ("CoolList<%s>::pop(): No elements in CoolList.\n", Type);
  abort ();
}


// remove_error -- Raise exception for CoolList::remove() method
// Input:          Type string
// Output:         None

void CoolList::remove_error (const char* Type) {
  //RAISE Error, SYM(CoolList), SYM(Invalid_Cpos),
  printf ("CoolList<%s>::remove(): Invalid current position.\n", Type);
  abort ();
}

// va_arg_error -- Raise exception for using class objects, or chars in (...)
// Input:          Type string
// Output:         None

void CoolList::va_arg_error (const char* Type, int n) {
  //RAISE Error, SYM(CoolList), SYM(Invalid_Va_Arg),
  printf ("CoolList<%s>::CoolList<%s>(): Invalid type in ... or wrong alignment with %d bytes.\n",
	  Type, Type, n);
  abort ();
}


